/***************************************************************************************/
/*                                                                                     */
/*           RIP protocol : server program                                             */
/*           Hardware handling of the matrix multiplication                            */
/*                                                                                     */
/***************************************************************************************/
/*                                                                                     */
/*           Pierre Rousseau and Carlos Aguilar Melchor.                               */
/*           Limoges University, France.                                               */
/*           All rights reserved.                                                      */
/*           Note: once our work is finished all this work will be under a GPL licence.*/
/*           Until then, we keep the mention "all rights reserved" for scientific      */
/*           conference publication issues... Sorry.                                   */
/*                                                                                     */
/***************************************************************************************/



#include <stdio.h>
#include <cutil.h>



typedef long long int64;


//#define _DEBUG

#define LINE_THREADS    2

#define N   50

#define NUM_ROWS_QUERY  N
#define NUM_COLS_QUERY  (2*N)


// definition of the query matrix, stored in constant memory.
// this cannot be defined inside a function.
__constant__ int64 device_query[NUM_ROWS_QUERY * NUM_COLS_QUERY];


// global variables, to avoid reallocating memory for each iteration
unsigned int size_data;
unsigned int mem_size_data;
unsigned short *device_data;

unsigned int size_query;
unsigned int mem_size_query;

unsigned int size_result;
unsigned int mem_size_result;
int64* device_partial_result;   // the result of the multiplication
int64* device_final_result;     // the sum of the successive multiplications



void initComputation(int num_rows_d, int num_cols_d, int num_rows_q, int num_cols_q)
{
    CUT_DEVICE_INIT();
    
    // memory allocation 
    
    // allocate device memory for the data matrix
    size_data = num_rows_d * num_cols_d;
    mem_size_data = sizeof(unsigned short) * size_data;
    CUDA_SAFE_CALL(cudaMalloc((void**) &device_data, mem_size_data));
    
    // device memory for the query matrix is already allocated (in constant memory)
    size_query = num_rows_q * num_cols_q;
    mem_size_query = sizeof(int64) * size_query;
    
    //allocate device memory for the result
    size_result = num_rows_d * num_cols_q;
    mem_size_result = sizeof(int64) * size_result;
    CUDA_SAFE_CALL(cudaMalloc((void**) &device_partial_result, mem_size_result));
    CUDA_SAFE_CALL(cudaMalloc((void**) &device_final_result, mem_size_result));
}



void endComputation()
{
    // release memory
    
    CUDA_SAFE_CALL(cudaFree(device_data));
    
    // device query doesn't have to be freed, as it is stored in constant memory
    // CUDA_SAFE_CALL(cudaFree(device_query));
    
    CUDA_SAFE_CALL(cudaFree(device_partial_result));
    CUDA_SAFE_CALL(cudaFree(device_final_result));
}






__global__ void kernelMult(unsigned short *data, int64 *result, int64 module)
{    
    // Block index
    int bx = blockIdx.x;
    //int by = blockIdx.y;

    // Thread index
    int tx = threadIdx.x;
    int ty = threadIdx.y;
    
    int dataBegin = (bx * LINE_THREADS + ty ) * N;         // data has N columns
    int resultBegin = (bx * LINE_THREADS + ty ) * 2 * N;   // result has 2N columns
    
    // load a chunk from the data matrix into a shared array :
    __shared__ int64 dataShared[LINE_THREADS][N];
    
    /*
    if (tx < N)
    {
        dataShared[ty][tx] = data[dataBegin + tx];
    }
    */
    
    if (tx % 2 == 0)
    {
        dataShared[ty][tx/2] = data[dataBegin + tx/2];
    }
    
    
    __syncthreads();

    // each thread will iterate, initialize the partial result
    int64 res = 0;
    
    for(int i = 0 ; i < N ; i++)
    {
        res = (res + dataShared[ty][i] * device_query[(i*2*N) + tx] ) % module;
    }
    result[resultBegin + tx] = res;
}






void GPUMatrixMul(unsigned short *d, int64 *q, int64 *r, int64 module, int num_rows_d, int num_cols_d, int num_rows_q, int num_cols_q)
{
    // setup matrix discretization
    dim3 grid(num_rows_d / LINE_THREADS, 1);   // divide result matrix : each block will consist in one line
    dim3 threads(num_cols_q, LINE_THREADS);   // divide each block : t.x will browse through result columns
    
    // copy host memory to device
    CUDA_SAFE_CALL( cudaMemcpy(device_data, d, mem_size_data, cudaMemcpyHostToDevice) );

    // upload query matrix to the device.
    CUDA_SAFE_CALL( cudaMemcpyToSymbol(device_query, q, mem_size_query, 0) );

    // invoke kernel
    kernelMult<<< grid, threads >>>(device_data, device_partial_result, module);
    
    CUT_CHECK_ERROR("Kernel execution failed");

    // copy result from device to host
    // not needed if we use the GPU add
    //CUDA_SAFE_CALL( cudaMemcpy(r, device_partial_result, mem_size_result, cudaMemcpyDeviceToHost) );
}





__global__ void kernelAdd(int64 *to, int64 *from, int64 module)
{    
    // Block index
    int bx = blockIdx.x;
    //int by = blockIdx.y;

    // Thread index
    int tx = threadIdx.x;
    int ty = threadIdx.y;
    
    int resultBegin = (bx * LINE_THREADS + ty ) * 2 * N;   // result has 2N columns
    
    int64 res = (to[resultBegin + tx] + from[resultBegin + tx]);
    to[resultBegin + tx] = (res>=module) ? (res-module) : (res) ;
}




void GPUMatrixAdd(int64 *a, int64 *b, int64 module, int num_rows, int num_cols)
{
    // setup matrix discretization
    dim3 grid(num_rows / LINE_THREADS, 1);   // divide result matrix : each block will consist in one line
    dim3 threads(num_cols, LINE_THREADS);   // divide each block : t.x will browse through result columns
    
    // copy host memory to device (final result)
    //CUDA_SAFE_CALL( cudaMemcpy(device_final_result, a, mem_size_result, cudaMemcpyHostToDevice) );

    // partial result is already in GPU mem
    
    // invoke kernel
    kernelAdd<<< grid, threads >>>(device_final_result, device_partial_result, module);

    // copy final result from device to host
    //CUDA_SAFE_CALL( cudaMemcpy(a, device_final_result, mem_size_result, cudaMemcpyDeviceToHost) );
}


void uploadResult(int64 *tmpResult)
{
    // copy host memory to device (final result)
    CUDA_SAFE_CALL( cudaMemcpy(device_final_result, tmpResult, mem_size_result, cudaMemcpyHostToDevice) );
}



void retrieveResult(int64 *tmpResult)
{
    // copy final result from device to host
    CUDA_SAFE_CALL( cudaMemcpy(tmpResult, device_final_result, mem_size_result, cudaMemcpyDeviceToHost) );
}


