/***************************************************************************************/
/*                                                                                     */
/*           RIP protocol : server program                                             */
/*                                                                                     */
/***************************************************************************************/
/***************************************************************************************/
/*                                                                                     */
/*           Pierre Rousseau and Carlos Aguilar Melchor.                               */
/*           Limoges University, France.                                               */
/*           All rights reserved.                                                      */
/*           Note: once our work is finished all this work will be under a GPL licence.*/
/*           Until then, we keep the mention "all rights reserved" for scientific      */
/*           conference publication issues... Sorry.                                   */
/*                                                                                     */
/***************************************************************************************/



// network includes
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>

// for the timer
#include <sys/time.h>

// standard includes
#include <fstream>
#include <string> 
#include <iostream>
#include <vector>
#include <cmath>
using namespace std;

// to close the program nicely with ctrl-C
#include <signal.h>

#include "CommonRIP.h"





//////////////////////////////////////////////////////////////
/// Predefinition of the CUDA functions
//////////////////////////////////////////////////////////////


// computation is done on GPU if the following directive is present, on CPU otherwise.
#define GPU_COMPUTATION


void initComputation(int num_rows_d, int num_cols_d, int num_rows_q, int num_cols_q);
void endComputation();
void GPUMatrixMul(unsigned short *d, int64 *q, int64 *r, int64 module, int num_rows_d, int num_cols_d, int num_rows_q, int num_cols_q);
void GPUMatrixAdd(int64 *a, int64 *b, int64 module, int num_rows, int num_cols);
void uploadResult(int64 *tmpResult);
void retrieveResult(int64 *tmpResult);






//////////////////////////////////////////////////////////////
/// This class prepares the file list for transmission to the client
/// It will send file list, and fileCount to the client.
//////////////////////////////////////////////////////////////
class DESC
{
    public :
        string configFile;
        vector<string> fileList;
        int fileCount;
        long maxSize;
        int chunkCount;
        int numIterationsPerChunk;
        int dataElementAtOnce;
        
        
        // constructor
        DESC (string _configFile)
        {
            configFile = _configFile;
        
            ifstream f( configFile.c_str() );
            int tmpSize=0;
            string line;
            
            maxSize = 0;
            fileCount = 0;
            
            if (f)
            {
                while( getline(f, line) )
                {
                    // beware of windows line encoding :
                    if (line[line.size() -1] == '\r')
                    {
                        // remove this ugly character !
                        line.erase(--line.end());
                    }
                    
                    tmpSize = getFileSize(line);
                    if (tmpSize > 0)
                    {
                        fileList.push_back(line);
                        
                        fileCount++;
                        if (tmpSize > maxSize)
                        {
                            maxSize = tmpSize;
                        }
                    }
                }
                cout << "Parsing of the config file done." << endl;
                f.close();
            }
            else
            {
                cout << "Problem opening config file." << endl;
            }
            
            // we have MAX_HOST_MEMORY_FOR_DATA to share at once between the various data files.
            // at a time, each file should get :
            // int dataElementAtOnce = MAX_HOST_MEMORY_FOR_DATA / fileCount;
            // Well, obviously this would be too easy, because it is better if we can avoid loading partial rows, as well as avoid having too much padding at the end of the files ...
            
            int minimum_loadable = N * LINES_COMPUTED_AT_ONCE * sizeof(unsigned short);
            
            // we will load k * (N columns X 800 rows)
            // how many chunks do we need ?
            int maximum_loadable = (int)ceil ( (MAX_HOST_MEMORY_FOR_DATA / float(fileCount) ) / float(minimum_loadable) ) * minimum_loadable;
            
            int maxSizeWithInitialPadding = maxSize + 8; // this takes into account the file size which is sent before the actual file.
            
            if(maxSizeWithInitialPadding < maximum_loadable)
            {
                // We won't have to load successie parts of the file, all in one go is OK
                numIterationsPerChunk = int(ceil(maxSizeWithInitialPadding / float(minimum_loadable)));    // number of iterations needed to process one chunk of data (one file read)
                dataElementAtOnce = numIterationsPerChunk * minimum_loadable;            // size each file will occupy in memory at once
                chunkCount = 1;                                                          // on chunk is enough in this case : we can load the whole file into memory at the beginning.
            }
            else
            {
                // We'll have to load a chunk of the file, process it, then load another, etc.
                numIterationsPerChunk = maximum_loadable / minimum_loadable;
                dataElementAtOnce = maximum_loadable;
                chunkCount = (int)ceil((float)maxSizeWithInitialPadding/(float)dataElementAtOnce);
                cout << "DATA LOADING IS HEAVILY UNOPTIMISED !!" << endl;
            }
            
            //int totalNumIterations = int(ceil(maxSizeWithInitialPadding / float(minimum_loadable)));    // total number of iterations needed to process the largest file, i.e. all files
            
            
            
            cout << "######################################################################" << endl;
            cout << "## Total data treated (including padding) : " << dataElementAtOnce * chunkCount * fileCount * ( 8 / (1024.0*1024.0) ) << "Mbits." << endl;
            cout << "######################################################################" << endl;
            //cout << "Min : " << minimum_loadable << " max : " << maximum_loadable << " maxsize : " << maxSize << " iterations : " << numIterationsPerChunk << " dataAtOnce : " << dataElementAtOnce << " chunkcount : " << chunkCount << endl;
        }
        
        
        ~DESC()
        {
            fileList.clear();
        }
        
        
        // function to get the size of a file, with filename as argument.
        long getFileSize(const string & fileName)
        {
            ifstream f( fileName.c_str() );
            
            if (f)
            {
                long pos = f.tellg();
                f.seekg( 0, std::ios_base::end );
                long size = f.tellg() ;
                f.seekg( pos, std::ios_base::beg ) ;
                
                f.close();
                
                return size ;
            }
            else
            {
                // file probably doesn't exist
                cout << "Invalid file : " << fileName << endl << "Please check your config file, and relaunch the program." << endl << endl;
                return 0;
            }
        }
        
        string getFileName(int index)
        {
            return fileList[index];
        }
        
        int getFileSize(int index)
        {
            return getFileSize(fileList[index]);
        }
        
        string dissociateFromPath(string fullPath)
        {
            int slashPosition = fullPath.find_last_of('/');
            if (slashPosition == -1)
            {
                return fullPath;
            }
            else
            {
                return fullPath.substr(slashPosition+1, fullPath.length()-slashPosition);
            }
        }
        
        void formatForTransmission(char *buffer)
        {            
            // buffer is supposed to have enough free space for transmission
            // first information : fileCount, rows of each result, cols of each result, number of results
            sprintf(buffer, "%d %d %d %d\n", fileCount, LINES_COMPUTED_AT_ONCE, 2 * N, chunkCount * numIterationsPerChunk);
            
            for (vector<string>::iterator it = fileList.begin() ; it < fileList.end() ; it++)
            {
                strcat(buffer, dissociateFromPath(*it).c_str());
                strcat(buffer, "\n");
            }
        }
};





//////////////////////////////////////////////////////////////
/// This class is a sub-matrix (data element) of the RIP data matrix
//////////////////////////////////////////////////////////////
class RIPmatrixData_i
{
    public :
        unsigned short *A_i;
        long maxSize, fileSize;
        long loadableSize;  // maximum size which can be loaded at once (depends on the number of elements in the database)
        string fileName;
        
        // constructor
        RIPmatrixData_i(long _maxSize, long _loadableSize, const string & _fileName, int _fileSize)
        {
            maxSize = _maxSize + 8; // 8 ==> filesize uses 64 bits
            loadableSize = _loadableSize;
            fileName = _fileName;
            
            A_i = (unsigned short *)malloc( loadableSize );
            
            fileSize = _fileSize;
            
            loadFile(0);
        }
        
        ~RIPmatrixData_i()
        {
            free(A_i);
        }
        
        // load of a data file as matrix element
        // shifting allows the load of parts of a file.
        bool loadFile(long shifting)
        {
            long memPos = 0; // refers to the memory position
            long fPos = 0;   // refers to the current position in the read file
            unsigned short *tmp = A_i;
            
            // open the file
            std::ifstream f(fileName.c_str(), std::ios_base::binary);
            
            if (shifting == 0)
            {
                cout << "Loading file : " << fileName << endl;
                
                int64 fs = fileSize;
                
                // load fileSize into A_i
                *tmp++ = fs>>48;
                *tmp++ = (fs<<16)>>48;
                *tmp++ = (fs<<32)>>48;
                *tmp++ = (fs<<48)>>48;
                
                memPos = 8; // already stored 8 bytes.
                fPos = 0;
                
                //cout << "size encoded : " << fs << endl;
            }
            else
            {
                cout << "Continuing file : " << fileName << ", chunk number " << (shifting/loadableSize) << endl;
                memPos = 0;
                fPos = shifting -8; // first 8 bytes are the filesize, not actual file data
                
                // already read some parts of the file => skip them
                if (fPos < fileSize)
                {
                    f.seekg(fPos, std::ios_base::beg);
                }
            }
            
            
            // now we can read the file
            while (fPos < fileSize && memPos < loadableSize)
            {   
                f.read(reinterpret_cast<char*>(tmp), sizeof(unsigned short));
                tmp++;
                memPos += 2;
                fPos += 2; // move two times, because a short is worth two bytes
            }
            f.close();
            
            //put some padding (if needed)
            if (memPos < loadableSize)
            {
                while (memPos < loadableSize)
                {
                    *tmp++ = 0;
                    memPos += 2;
                }
            }
            
            return true;
        }
        
        
        // write function, from memory to file, for debug purpose
        bool writeFile(string fileName)
        {
            std::fstream f(fileName.c_str(), std::ios_base::out | std::ios_base::binary | std::ios_base::app);
            long pos = 0;
            unsigned short *tmp = A_i;
            
            if (f)
            {
                
                //long pos = f.tellp();
                f.seekp( 0, std::ios_base::end );
                long alreadyWritten = f.tellp() ;
                
                if (alreadyWritten == 0)
                {
                    // we are beginning to write the file.
                    
                    // skip filesize, stored at the beginning of the matrix
                    
                    int64 storedFileSize = 0;
                    int64 i;
                    
                    i = (unsigned short)(*tmp++);
                    storedFileSize += i<<48;
                    i = (unsigned short)(*tmp++);
                    storedFileSize += i<<32;
                    i = (unsigned short)(*tmp++);
                    storedFileSize += i<<16;
                    i = (unsigned short)(*tmp++);
                    storedFileSize += i;
                    
                    cout << "size decoded : " << storedFileSize << endl;
                    
                    //some data has already been extracted from the matrix
                    pos = 8;
                }
                
                while (alreadyWritten < fileSize && pos < loadableSize)
                {
                    f.write(reinterpret_cast<const char*>(tmp++), sizeof(unsigned short));
                    pos += 2; // move two times, because a short is worth two bytes
                    alreadyWritten += 2;
                    
                    //using shorts might be a problem in some cases (text files with odd character count for example..)
                    if (alreadyWritten == fileSize -1)
                    {
                        f.write(reinterpret_cast<const char*>(tmp++), sizeof(char));
                        alreadyWritten++;
                        pos++;
                    }
                }
                f.close();
                
                cout << "Already written : " << alreadyWritten/(1024.0*1024.0) << " Mo / " << fileSize/(1024.0*1024.0) << " Mo (" << alreadyWritten * 100.0 / float(fileSize) << "%)." << endl;
                
                return true;
            }
            else
            {
                cout << "Couldn't open output file for writing." << endl;
                return false;
            }
        }
};







//////////////////////////////////////////////////////////////
/// This class holds a full RIP data matrix
//////////////////////////////////////////////////////////////
class RIPmatrixData
{
    public :
        vector<RIPmatrixData_i *> A;
        long readSoFar;
        DESC *catalogue;
        
        // constructor
        RIPmatrixData(DESC *_cat)
        {
            catalogue = _cat;
            
            readSoFar = 0;
            A.clear();
        }
        
        // destructor
        ~RIPmatrixData()
        {
            // suppression des sous-matrices data_i
            for (vector<RIPmatrixData_i *>::iterator it = A.begin() ; it < A.end() ; it++)
            {
                delete(*it);
                *it = NULL;
            }
            A.clear();
        }
        
        // function which loads the first part of a file
        bool loadFirst()
        {        
            string fileName;
            int fileSize;
            
            for (int i = 0 ; i < catalogue->fileCount ; i++)
            {    
                fileName = catalogue->getFileName(i);
                fileSize = catalogue->getFileSize(i);
                
                if (fileSize != 0)
                {
                    A.push_back(new RIPmatrixData_i(catalogue->maxSize, catalogue->dataElementAtOnce, string(fileName), fileSize));
                }
            }
            
            readSoFar = catalogue->dataElementAtOnce;
            return true;
        }
        
        
        // load function, to be called only after 'loadFirst' has been called once
        bool loadNext()
        {
            // uncomment if you wish to test if write works ok
            //A[0]->writeFile(string("glopiglopa"));
            
            if (readSoFar < catalogue->maxSize)
            {
                // there is more to be read from the files
                
                for (vector<RIPmatrixData_i *>::iterator it = A.begin() ; it < A.end() ; it++ )
                {
                    (*it)->loadFile(readSoFar);
                }
                readSoFar += catalogue->dataElementAtOnce;
                return true;
            }
            else
            {
                return false;
            }
        }
};





class RIPmatrixQuery
{
    public :
        int num_elems;
        vector<int64 *> M_i_list; //this is the query !
        int64 module;
        
        RIPmatrixQuery(int _num_elems)
        {
            num_elems = _num_elems;
        }
        
        
        ~RIPmatrixQuery()
        {
            for (vector<int64 *>::iterator it = M_i_list.begin() ; it < M_i_list.end() ; it++)
            {
                free(*it);
            }
            M_i_list.clear();
        }
        
        
        void setQuery(int64 *q)
        {
            M_i_list.push_back(q);
        }
        
        void setModule(int64 mod)
        {
            module = mod;
        }
        
};



class RIPServer
{
    public :

        int mainSocket, connectionSocket; // connectionSocket is set after an 'accept' call
        int domain, type, protocol;
        sockaddr_in source_info;
        struct sockaddr cliaddr;
        
        // constructor
        RIPServer()
        {
            int error;
            
            cout << "Initiating server." << endl;
            domain = AF_INET;
            type = SOCK_STREAM;
            protocol = 0;
            
            // socket creation
            mainSocket = socket (domain, type, protocol);
            
            // socket binding
            source_info.sin_family=AF_INET;
            source_info.sin_addr.s_addr=INADDR_ANY; // listen to all local IP's
            source_info.sin_port=htons(SERVER_PORT); // the port we're listening to
            error=bind(mainSocket,(struct sockaddr*)&source_info,sizeof(source_info));
            if (error!=0)
            {
                cout << "Couldn't bind server socket." << endl;
                exit (2);
            }
            
            // socket listening
            error = 99;
            error = listen(mainSocket,1);
            if (error != 0)
            {
                cout << "Problem listening to socket, aborting." << endl;
                exit (3);
            }
            else
            {
                // OK, connections can be received now
            }
        }
        
        
        // destructor : close socket
        ~RIPServer()
        {
            shutdown(mainSocket, 2);
            close (mainSocket);
        }
        
        
        // connection handler
        int getConnection()
        {
            socklen_t addrlen = sizeof(source_info);
            
            cout << "Waiting for connections..." << endl;

            connectionSocket = accept(mainSocket, (struct sockaddr*)&cliaddr, &addrlen); // returns the new socket id.
            
            // when we reach this point, a connection has been accepted
            cout << "Connection received !" << endl;
            return connectionSocket;
        }
        
        
        void closeConnectionSocket(int s)
        {
            shutdown(s, 2);
            close (s);
        }
        
        
        
        bool getQuery(RIPmatrixQuery *q)
        {
            int bufferSize = N * 2*N * sizeof(int64);
            char *buffer = (char*)malloc(bufferSize);
            
            cout << "Waiting for query." << endl;
            
            // first : read the modulus to use (p)
            char moduleBuffer[8];
            bzero(moduleBuffer, sizeof(moduleBuffer));
            recv(connectionSocket, moduleBuffer, sizeof(moduleBuffer), 0);
            int64 mod;
            memcpy(&mod, moduleBuffer, 8);
            q->setModule(mod);
            
            //cout << "Module : " << mod << endl;
            
            // and now, get the actual data
            for (int i = 0 ; i < q->num_elems ; i++)
            {
                int64 *query = (int64 *)malloc (bufferSize);

                int charactersRead = 0;
                
                bzero(buffer, bufferSize);
                
                // we can't download all at once
                // let's split it in sections of 1k
                int downloaded = 0;
                int downloadSize = 1024;
                char *tmpBuf = buffer;
                
                while (downloaded != bufferSize)
                {
                    int nextDownload = (bufferSize - downloaded)<downloadSize ? (bufferSize - downloaded) : downloadSize;
                
                    charactersRead=recv(connectionSocket, tmpBuf, nextDownload, 0);
                    downloaded += charactersRead;
                    tmpBuf += charactersRead; // pointer shifting, so as not to override the just written data in the buffer.
                    if (charactersRead != downloadSize && downloaded != bufferSize)
                    {
                        cout << "I didn't get everything I expected : " << downloaded << " instead of " << bufferSize << endl;
                        return false;
                    }
                }
                
                
                memcpy(query, buffer, bufferSize);
                q->setQuery(query);
            }
            
            // clean the mess behind us !
            free(buffer);
            
            cout << "Query received OK." << endl;
            return true;
        }
        
        
        bool uploadResult(int64 *tmpResult, int num_rows, int num_cols)
        {
            int bufferSize = sizeof(int64) * num_rows * num_cols;
            char *buffer = (char *)malloc(bufferSize);
            
            //cout << "Uploading partial result (" << (num_rows * num_cols) << " elements)." << endl;
            
            bzero(buffer, sizeof(buffer));
            memcpy(buffer, tmpResult, num_rows * num_cols * sizeof(int64));
            
            // we can't upload all at once
            // let's split it in sections of 1k
            int uploaded = 0;
            int uploadSize = 1024;
            while (uploaded != bufferSize)
            {
                uploadSize = (uploaded + uploadSize) < bufferSize ? uploadSize : (bufferSize - uploaded);
                
                int charactersWritten=send(connectionSocket, buffer+uploaded, uploadSize, 0);
                uploaded += charactersWritten;
                if (charactersWritten != uploadSize && uploaded != bufferSize)
                {
                    cout << "The whole answer wasn't properly uploaded to the client :(" << endl;
                    return false;
                }
            }
            
            //cout << "Uploaded !" << endl;
            
            free(buffer);
            
            return true;
        }
};





class RIPmatrixResult
{
    public :
        RIPmatrixData * data;
        RIPmatrixQuery * query;
        RIPServer * serv;
        
        RIPmatrixResult(RIPmatrixData * _data, RIPmatrixQuery * _query, RIPServer *_serv)
        {
            data = _data;
            query = _query;
            serv = _serv;
            
#ifdef GPU_COMPUTATION
            // CUDA initialization
            initComputation(LINES_COMPUTED_AT_ONCE, N, N, 2*N);
#endif

            cout << "Beginning result computation." << endl;
            
            // allocate memory for the result
            int64 *tmpResult = (int64 *)malloc(sizeof(int64) * (2*N) * LINES_COMPUTED_AT_ONCE);
            
            // allocate memory for the -- part(A_i) * Bi -- multiplication
            int64 *tmpPartialResult = (int64 *)malloc(sizeof(int64) * (2*N) * LINES_COMPUTED_AT_ONCE);
            
            // set a timer
            struct timeval t0;
            gettimeofday(&t0, NULL);
            
            int stepper = 0;
            
            do
            {                
                // process current data
                // current data won't be loaded all at once in video memory. We have a few iterations to do.
                
                for (int iter=0 ; iter < data->catalogue->numIterationsPerChunk ; iter++)
                {
                    cout << "Iteration " << ++stepper << "/" << (data->catalogue->numIterationsPerChunk * data->catalogue->chunkCount) << endl;
                    
                    // initialize the result to 0
                    bzero(tmpResult, sizeof(int64)* LINES_COMPUTED_AT_ONCE * (2*N) );
#ifdef GPU_COMPUTATION
                    uploadResult(tmpResult);
#endif
                    
                    for (int i = 0 ; i < query->num_elems ; i++)
                    {
                        unsigned short *tmpData = data->A[i]->A_i;
                        
                        // move tmpData so it is correctly positionned for this iteration
                        tmpData += iter * N * LINES_COMPUTED_AT_ONCE;
                        
                        // initialize result to 0
                        bzero(tmpPartialResult, sizeof(int64)* LINES_COMPUTED_AT_ONCE * (2*N) );
                        
                        // make multiplication : this HAS to take the modulus into account. The addition also has to.
#ifdef GPU_COMPUTATION
                        GPUMatrixMul(tmpData, query->M_i_list[i], tmpPartialResult, query->module, LINES_COMPUTED_AT_ONCE, N, N, 2*N);
                        GPUMatrixAdd(tmpResult, tmpPartialResult, query->module, LINES_COMPUTED_AT_ONCE, 2*N);
#else
                        CPUMatrixMul(tmpData, query->M_i_list[i], tmpPartialResult, query->module, LINES_COMPUTED_AT_ONCE, N, N, 2*N);
                        CPUMatrixAdd(tmpResult, tmpPartialResult, query->module, LINES_COMPUTED_AT_ONCE, 2*N);
#endif
                    }
                    
#ifdef GPU_COMPUTATION
                    retrieveResult(tmpResult);
#endif
                    // OK, now we have a partial result to upload to the client :)
                    serv->uploadResult(tmpResult, LINES_COMPUTED_AT_ONCE, 2*N);
                }
                
                // next step of computation
            }
            while (data->loadNext());
            
            // so, how long was it ?
            struct timeval t1;
            gettimeofday(&t1, NULL);
            float t = (t1.tv_sec - t0.tv_sec) + (t1.tv_usec - t0.tv_usec) * 0.000001f;
            
            cout << "######################################################################" << endl;
            cout << "## Total data treated (including padding) : " << data->catalogue->dataElementAtOnce * data->catalogue->chunkCount * data->catalogue->fileCount * ( 8 / (1024.0*1024.0) ) << "Mbits." << endl;
            cout << "######################################################################" << endl;
            cout << "## Total computation time : " << t << " secs." << endl;
            cout << "######################################################################" << endl;
            cout << "## Data rate : " << data->catalogue->dataElementAtOnce * data->catalogue->chunkCount * data->catalogue->fileCount * ( 8 / (1024.0*1024.0) ) / t << " Mbits/secs." << endl;
            cout << "######################################################################" << endl;
            
            // clean up !
            
#ifdef GPU_COMPUTATION
            // CUDA cleaning
            endComputation();
#endif

            free(tmpResult);
            free(tmpPartialResult);
        }
        
        
        ~RIPmatrixResult()
        {
            cout << "Destroying result matrix." << endl;
        }
        
        
        void CPUMatrixMul(unsigned short *d, int64 *q, int64 *r, int64 module, int num_rows_d, int num_cols_d, int num_rows_q, int num_cols_q)
        {
            if (num_cols_d != num_rows_q)
            {
                cout << "Impossible Matrixmul : size mismatch." << endl;
            }
            else
            {
                for (int l = 0 ; l < num_rows_d ; l++)
                {
                    for (int c = 0 ; c < num_cols_q ; c++)
                    {
                        for (int elem = 0 ; elem < num_rows_q ; elem++)
                        {
                            // it shouldn't be necessary to compute two modulus, one seems to be enough.
                            //r[l*num_cols_q + c] = (r[l*num_cols_q + c] + (d[l*num_cols_d + elem] * q[elem*num_cols_q + c])%module)%module ;
                            
                            r[l*num_cols_q + c] = (r[l*num_cols_q + c] + (d[l*num_cols_d + elem] * q[elem*num_cols_q + c]))%module ;
                        }
                    }
                }
            }
        }
        
        
        void CPUMatrixAdd(int64 *a, int64 *b, int64 module, int num_rows, int num_cols)
        {
            int64 *tmpA = a;
            int64 *tmpB = b;
            
            for (int i = 0 ; i < num_rows*num_cols ; i++)
            {
                // we add elements in the same positions, so don't have to use the array as two dimensional
                
                // as in the multiplication, one modulus should be enough. Nevertheless, this one must be done as long as we sul more than 16 elements (i.e. database contains more than 16 files).
                //int64 intA = *tmpA % module;
                //int64 intB = *tmpB % module;
                //int64 res = (intA + intB) % module;
                
                int64 res = (*tmpA + *tmpB) % module;
                
                *tmpA = res;
                tmpA++;
                tmpB++;
            }
        }
};







/////////////////////////////////////////////////////////
//      entry point to the program
/////////////////////////////////////////////////////////
int main (int argc, char **argv)
{    
    if (argc != 2)
    {
        cout << "This Program requires one argument (config file, listing the files in the database). Please call it properly." << endl;
    }
    else
    {   
        string configFile = argv[1];
        
        int bufferSize = 65535;
        char *buffer = (char*)malloc(bufferSize);
        int activeSocket;
        
        RIPmatrixData *data;
        RIPmatrixQuery *query;
        RIPmatrixResult *V;
        
        RIPServer *serv;
        
        
        DESC *catalogue = new DESC(configFile);
        
        
        serv = new RIPServer();

        
        // infinite loop to keep the server up and running when a client is done
        
        //while (true)
        {
            data = new RIPmatrixData(catalogue);
            data->loadFirst();

            activeSocket = serv->getConnection();
            
            // get first message
            
            // make sure the buffer is all zeros, then read the message from the client
            bzero(buffer, bufferSize);
            recv(activeSocket,buffer,bufferSize,0);
            
            if (strcmp(buffer, "Hello") != 0)
            {
                query = NULL;
                V = NULL;
                
                // connection messed up
                cout << "Bad connection. Asking the client to retry." << endl;
                
                strcpy(buffer,"Retry"); // prepare buffer for transmission to server.
                send(activeSocket,buffer,strlen(buffer),0);
            }
            else
            {
                // client has introduced himself, we shall send him the file list.
                cout << "Connection OK. Sending file list." << endl;
                
                catalogue->formatForTransmission(buffer);
                // send DESC file
                send(activeSocket, buffer,strlen(buffer),0);
                
                // now we can receive the request !
                query = new RIPmatrixQuery(catalogue->fileCount);
                
                if (serv->getQuery(query))
                {
                    // we got the query, now we can compute the result matrix.
                    V = new RIPmatrixResult(data, query, serv);
                }
                else
                {
                    V = NULL;
                }
            }
            
            // end connection and get ready for a new one.
            serv->closeConnectionSocket(activeSocket);
            
            //get ready for next connection :
            delete(V);
            delete(query);
            delete(data);
            
            cout << "data deleted" << endl;
        }
        
        free(buffer);
        delete(serv);
        delete(catalogue);
    }
    
    
    cout << "Bye bye !" << endl;
}




